"""
Binho Pulsar Python package

This Python package is developed to control the new Pulsar USB Host adapter.
The new Binho Pulsar device is an USB HID device that is controlled using a set
of USB commands that are transferred over USB HID Interrupt transfers.
"""
# Import drivers.
from .supernovasdk.binhosupernova.supernova import Supernova

# Import USB HID commands library.
from .commands.gpio.definitions import *
from .commands.i2c.definitions import *
from .commands.spi.definitions import *
from .commands.uart.definitions import *
from .commands.system.definitions import *

# Import other modules
from .supernovasdk.binhosupernova.utils.system_message import *

class Pulsar():
    """
    This class represents the Binho Pulsar device.

    It involves the drivers used to communicate with the device over USB HID,
    and the different processes that control the commands request and responses.
    """

    def __init__(self):
        """
        Constructor defined to add the Pulsar attributes and initialize them as None.
        """
        self.__driver = Supernova() # We utilize the supernova class as low level driver, since its API and pulsars is very similar

    # This function tries to hotfix a hardcoded string used in a wrapper in the Supernova class.
    def __error_message_parser_wrapper(func):
        def __check_error_message(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            if result["opcode"] == SystemOpcode.CONFIGURATION_ERROR.value:
                result["message"] = "Pulsar not configured appropriately, please check if connection is open and callback is set"
            return result
        return __check_error_message

    # Public methods ----------------------------------------------------------------------------------

    def open(self, serial: int = None, path: str = None) -> dict:
        """
        This method establishes the USB communication with the Pulsar. When this method is invoked, the USB Host starts to send USB frames to the
        USB device, and to poll the Input endpoint.

        By default, the usb hid manager, opened connection with the first Pulsar listed based on the
        VID and PID. Optionally, the serial number or the OS path string can be used to identify the
        device, so passing the serial number or the path only is enough to connect with the desired device.

        To get the device path or serial number, the BinhoPulsar.getConnectedPulsarDevicesList() method
        can be invoked.

        Parameters
        ----------
        serial : int
            USB device serial number. By default is None.
        
        path : str
            String path generated by the OS that can be read using the BinhoPulsar.getConnectedPulsarDevicesList() method.

        Returns
        -------
        dict:
            A python dictionary representation of a SystemMessage instance. The opcode of the message is:
            - SystemOpcode.OK if the connection with the HID device is completed successfully.
            - SystemOpcode.OPEN_CONNECTION_FAIL if the connection of the HID instance fails.

        """

        result = self.__driver.open(serial=serial, path=path)

        if result["opcode"] == SystemOpcode.OK.value:
            result["message"] = "Connection with Pulsar device opened successfully."

        return result

    def close(self) -> dict:
        """
        This method closes the communication with the Pulsar and releases the used memory.
        
        Returns
        -------
        dict
            A python dictionary representation of a SystemMessage instance. The opcode of the message is:
            - SystemOpcode.OK if the connection with the HID device is closed successfully.
            - SystemOpcode.OPEN_CONNECTION_REQUIRED if the was not previously opened.

        """

        result = self.__driver.close()

        if result["opcode"] == SystemOpcode.OPEN_CONNECTION_REQUIRED.value:
            result["message"] = "It is required to open connection with a Pulsar first. Invoke open() method."
        
        return result

    def onEvent(self, callback_func) -> dict:
        """
        This method registers the callback invoked every time a new USB response or notification
        arrives from the USB device. The callback function must implement the following signature:

            def callback(pulsar_response, system_message) -> None:

        It's important to note that the callback is called directly. If the callback function
        introduces a delay, it can impact the receiving mechanism. Developers using this SDK are
        advised to implement a queuing mechanism to quickly liberate the SDK as soon as the callback
        is called. This approach helps in managing responses efficiently without blocking the SDK's
        processing capabilities.

        Parameters
        ----------
        callback_func : function
            Callback function that will be invoked every time a new USB response or notification is
            sent by the USB device.

        Returns
        -------
        dict
            The SystemMessage in dictionary format.

        """

        def wrapped_callback_func(msg, sys):  
            if sys and sys.opcode == SystemOpcode.UNEXPECTED_DISCONNECTION:  
                sys.message =  f"Error {SystemOpcode.UNEXPECTED_DISCONNECTION.name}: Unexpected Pulsar disconnection."  
            callback_func(msg, sys)

        # this is to ensure the core callback is appropriate before setting.
        result = self.__driver.onEvent(callback_func=callback_func)  

        if result["opcode"] == SystemOpcode.INVALID_CALLBACK_SIGNATURE.value:  
            result["message"] = "The function must accept 2 Parameters: callback(pulsar_response, system_message)."  
            return result  

        # if core callback did not error, we can set the wrapper over it
        result = self.__driver.onEvent(callback_func=wrapped_callback_func)  
        
        return result

    # -----------------------------------------------------------------------------------
    # Communication API - USB commands
    # -----------------------------------------------------------------------------------

    # Get USB strings -------------------------------------------------------------------
    
    @__error_message_parser_wrapper
    def getUsbString(self, id: int, subCommand : GetUsbStringSubCommand) -> dict:

        """
        This function sends a Get USB String command taking the subcommand
        passed as parameter. The list of subcommands:

        GetUsbStringSubCommand.MANUFACTURER - Returns the manufacturer string ("Binho").
        GetUsbStringSubCommand.PRODUCT_NAME - Returns the product name string ("Binho Pulsar").
        GetUsbStringSubCommand.SERIAL_NUMBER - Returns the product serial number. Up to 32 characters.
        GetUsbStringSubCommand.FW_VERSION - Returns the product firmware version (VX.X.X).
        GetUsbStringSubCommand.HW_VERSION - Returns the product hardware version (HW-X).
        GetUsbStringSubCommand.BT_VERSION - Returns the product bootloader version (VX.X.X).

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].

        subCommand : GetUsbStringSubCommand
            Subcommand that indicates what string must be retrieved.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.getUsbString(id=id, subCommand=subCommand)

    @__error_message_parser_wrapper
    def resetDevice(self, id: int) -> dict:
        """
        This function sends a RESET DEVICE command.

        Parameters
        ----------
        id : c_uint16
            A 2-byte unsigned integer representing the transfer ID. The range allowed is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.  
        """
        return self.__driver.resetDevice(id=id)

    @__error_message_parser_wrapper
    def enterBootMode(self, id: int) -> dict:
        """
        This function sends a ENTER BOOT MODE command.

        Parameters
        ----------
        id : c_uint16
            A 2-byte unsigned integer representing the transfer ID. The range allowed is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.  
        """
        
        return self.__driver.enterBootMode(id=id)
    
    @__error_message_parser_wrapper
    def enterIspMode(self, id: int) -> dict:
        """
        This method sends a ENTER_ISP_MODE command

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """

        return self.__driver.enterIspMode(id=id)

    # Set I2C SPI UART Bus Voltage strings -------------------------------------------------------------------

    @__error_message_parser_wrapper
    def setI2cSpiUartGpioVoltage(self, id: int, voltage_mV : c_uint16) -> dict:
        """
        This method sends a SET_I2C_SPI_UART_GPIO_VOLTAGE command taking desired bus voltage

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        voltage_mV : c_uint16
            It is a 2-bytes integer that indicates the I2C, SPI, UART and GPIO operating voltage. The allowed range is [1200, 3300] mV
            and the value 0 mV to power off the output voltage.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.setI2cSpiUartGpioVoltage(id=id, voltage_mV=voltage_mV)

    @__error_message_parser_wrapper
    def getAnalogMeasurements(self, id: int) -> dict:
        """
        This method sends a GET_ANALOG_MEASUREMENTS command and retrieves the analog measures in mV.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.getAnalogMeasurements(id=id)

    @__error_message_parser_wrapper    
    def useExternalI2cSpiUartGpioVoltage(self, id: int) -> dict:
        """
        This method sends a USE_EXT_SRC_I2C_SPI_UART_GPIO_VOLTAGE command, sets the bus voltage to a measured value and
        retrieves the analog signal measure in mV.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.useExternalI2cSpiUartGpioVoltage(id=id)
    
    @__error_message_parser_wrapper
    def getDeviceInfo(self, id: int) -> dict:
        """
        This function sends a GET_DEVICE_INFO command to the device. The device returns a dictionary
        with its information.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.getDeviceInfo(id=id)

    # I2C management --------------------------------------------------------------------

    @__error_message_parser_wrapper
    def i2cControllerInit(self, id: int, busId: I2cBus, frequency: int, pullUpResistorsValue: I2cPullUpResistorsValue) -> dict:
        """
        This method initialize the I2C bus in controller mode with
        the frequency and pull up resistors value selected.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].
        
        busId : I2cBus
            The I2C bus identifier.

        frequency : int
            This parameter represents the I2C SCL frequency in Hz. Currently, the minimum allowed value
            is 100000 Hz and the maximum allowed value is 1000000 Hz.

        pullUpResistorsValue : I2cPullUpResistorsValue
            Value of the pull up enum that represents the desired Pull Up resistors value

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.i2cControllerInit(id=id, busId=busId, frequency=frequency, pullUpResistorsValue=pullUpResistorsValue)

    @__error_message_parser_wrapper
    def i2cControllerSetParameters(self, id: int, busId: I2cBus, frequency: int, pullUpResistorsValue: I2cPullUpResistorsValue) -> dict:
        """
        This method sets the I2C transfers baudrate and pull up resistors value for the SDA and SCL lines.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].
        
        busId : I2cBus
            The I2C bus identifier.

        frequency : int
            This parameter represents the I2C SCL frequency in Hz. Currently, the minimum allowed value
            is 100000 Hz and the maximum allowed value is 1000000 Hz.
        
        pullUpResistorsValue : I2cPullUpResistorsValue
            Value of the pull up enum that represents the desired Pull Up resistors value

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.i2cControllerSetParameters(id=id, busId=busId, frequency=frequency, pullUpResistorsValue=pullUpResistorsValue)

    @__error_message_parser_wrapper
    def i2cSetPullUpResistors(self, id: int, busId: I2cBus, pullUpResistorsValue: I2cPullUpResistorsValue) -> dict:
        """
        This method sets the I2C pull up resistors for the SDA and SCL lines.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].

        busId : I2cBus
            The I2C bus identifier.

        pullUpResistorsValue : I2cPullUpResistorsValue Enum
            This parameter represents the different values for the pull up resistors.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.i2cSetPullUpResistors(id=id, busId=busId, pullUpResistorsValue=pullUpResistorsValue)

    @__error_message_parser_wrapper
    def i2cControllerWrite(self, id: int, busId: I2cBus, targetAddress : int, registerAddress: list, data: list, isNonStop: bool = False, is10BitTargetAddress: bool = False) -> dict:
        """
        This method is used to request to the the Supernova device to perform an I2C write transfer. The
        I2C write transfer starts with a START condition and ends with a STOP condition. To perform a write
        transfer without the STOP condition, the nonStop parameter must be set to True.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        busId : I2cBus
            The I2C bus identifier.

        targetAddress : byte
            I2C slave static address.

        registerAddress : list
            Python list that contains the memory/register address of the I2C slave internal memory, whose data
            will be written. The list holds bytes, and can hand hold from 0 bytes up to 4 bytes. 0 bytes means
            that the list can be left empty and the Supernova will ignore it and write only the data payload.

        data : list
            Python list that contains the I2C data transferred in the I2C Write transfer. The list holds
            bytes elements, and the maximum length is 1024 bytes.
        
        isNonStop : bool
            This parameter indicates if the I2C write transfer is non-stop, meaning that it ends without a STOP 
            condition. If the flag is set to False, the transfer ends with STOP; otherwise, if it's True, the 
            transfer is non-stop, without a STOP condition at the end.

        is10BitTargetAddress : bool
            This parameter indicates if the target address is a 10-bit address. If the flag is set to False, the
            target address is a 7-bit address; otherwise, if it's True, the target address is a 10-bit address.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.i2cControllerWrite(id=id, busId=busId, targetAddress=targetAddress, registerAddress=registerAddress, data=data, isNonStop=isNonStop, is10BitTargetAddress=is10BitTargetAddress)

    @__error_message_parser_wrapper
    def i2cControllerRead(self, id: int, busId: I2cBus, targetAddress: int, requestDataLength: int, registerAddress: list = [], is10BitTargetAddress: bool = False) -> dict:
        """
        This method is used to request to the the USB device to perform an I2C read transfer. The
        I2C read transfer starts with a START condition and ends with a STOP condition.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        busId : I2cBus
            The I2C bus identifier.

        targetAddress : byte
            I2C target static address.

        requestDataLength : int
            Length of the read data. The maximum value is 1024 bytes.
        
        registerAddress : list
            Python list that contains the memory/register address of the I2C target internal memory, whose data
            will be read. The list holds bytes, and can hand hold from 0 bytes up to 4 bytes. 0 bytes means
            that the list can be left empty, the Supernova will ignore it and will perform only a read transfer.

        is10BitTargetAddress : bool
            This parameter indicates if the target address is a 10-bit address. If the flag is set to False, the
            target address is a 7-bit address; otherwise, if it's True, the target address is a 10-bit address.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """

        return self.__driver.i2cControllerRead(id=id, busId=busId, targetAddress=targetAddress, requestDataLength=requestDataLength, registerAddress=registerAddress, is10BitTargetAddress=is10BitTargetAddress)
    
    @__error_message_parser_wrapper
    def i2cControllerScanBus(self, id: int, busId: I2cBus, include10BitAddresses: bool = False) -> dict:
        """
        This method is used to request to the the USB device to perform an I2C scan of the I2C bus.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The allowed range is [1, 65535].

        busId : I2cBus
            The I2C bus identifier.
        
        include10BitAddresses : bool
            This parameter indicates if the scan must include 10-bit addresses. If the flag is set to False, the
            scan will not include 10-bit addresses; otherwise, if it's True, the scan will include 10-bit addresses.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """

        return self.__driver.i2cControllerScanBus(id=id, busId=busId, include10BitAddresses=include10BitAddresses)

    # --------------------------UART management --------------------------------------------------------------------

    # ------------------------------UART INIT ----------------------------#

    @__error_message_parser_wrapper
    def uartInit(self, id: int, baudrate: UartBaudRate, hardwareHandshake:bool, parityMode:UartParity, dataSize:UartDataSize, stopBit: UartStopBit ) -> dict:
        """
        This method initializes the UART peripherals

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        baudrate : UartBaudRate
            This parameter represents the UART TX and RX frequency from the options provided by the UartBaudRate enum.
            The frequency goes from 600bps to up to 115200bps.

        hardwareHandshake : bool
            This parameter represents a boolean flag to enable or disable this option.

        parityMode: UartParity
            This parameter represents the different parity modes available in the UartParity enum.
            The parity modes are: none, even or odd.

        dataSize: UartDataSize
            This parameter represents the different data sizes available in the UartDataSize enum.
            The data sizes are either 7 or 8.

        stopBit: UartStopBit
            This parameter represent the different stop bit configuration available in the UartStopBit enum.
            The stop bit can be of size 1 or 2.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """

        return self.__driver.uartInit(id=id, baudrate=baudrate, hardwareHandshake=hardwareHandshake, parityMode=parityMode, dataSize=dataSize, stopBit=stopBit)
    # ------------------------------UART SET ----------------------------#

    @__error_message_parser_wrapper
    def uartSetParameters(self, id: int, baudrate: UartBaudRate, hardwareHandshake:bool, parityMode:UartParity, dataSize:UartDataSize, stopBit: UartStopBit ) -> dict:
        """
        This method sets the UART peripheral parameters

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].

        baudrate : UartBaudRate
            This parameter represents the UART TX and RX frequency from the options provided by the UartBaudRate enum.
            The frequency goes from 600bps to up to 115200bps.

        hardwareHandshake : bool
            This parameter represents a boolean flag to enable or disable this option.

        parityMode: UartParity
            This parameter represents the different parity modes available in the UartParity enum.
            The parity modes are: none, even or odd.

        dataSize: UartDataSize
            This parameter represents the different data sizes available in the UartDataSize enum.
            The data sizes are either 7 or 8.

        stopBit: UartStopBit
            This parameter represent the different stop bit configuration available in the UartStopBit enum.
            The stop bit can be of size 1 or 2.


        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """

        return self.__driver.uartSetParameters(id=id, baudrate=baudrate, hardwareHandshake=hardwareHandshake, parityMode=parityMode, dataSize=dataSize, stopBit=stopBit)
    
    # ------------------------------UART SEND ----------------------------#

    @__error_message_parser_wrapper
    def uartSendMessage(self, id: int, data: list) -> dict:
        """
        This method is used to request to the the Pulsar device to perform an UART send transfer.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represent the transfer id. The range allowed is [1, 65535].

        data : list
            Python list that contains the data transferred in the UART Send. The list holds
            bytes elements, and the maximum length is 1024 bytes.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        
        return self.__driver.uartSendMessage(id=id, data=data)
    
    # SPI management --------------------------------------------------------------------

    @__error_message_parser_wrapper               
    def spiControllerInit(self,
                          id: int,
                          bitOrder: SpiControllerBitOrder,
                          mode: SpiControllerMode,
                          dataWidth: SpiControllerDataWidth,
                          chipSelect: SpiControllerChipSelect,
                          chipSelectPol: SpiControllerChipSelectPolarity,
                          frequency: c_uint32) -> dict:
        """
        This method initializes the SPI peripheral with the specified configuration.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        bitOrder : SpiControllerBitOrder
            Sets the bit order of the SPI peripheral, could be SpiControllerBitOrder.MSB or SpiControllerBitOrder.LSB.
        
        mode : SpiControllerMode
            Sets the SPI mode, could be SpiControllerMode.MODE_0, SpiControllerMode.MODE_1, SpiControllerMode.MODE_2 or SpiControllerMode.MODE_3.
        
        dataWidth : SpiControllerDataWidth
            Sets the SPI data width, could be SpiControllerDataWidth._8_BITS_DATA, SpiControllerDataWidth._16_BITS_DATA.

        chipSelect : SpiControllerChipSelect
            Sets the SPI chip select, could be SpiControllerChipSelect.CHIP_SELECT_0, SpiControllerChipSelect.CHIP_SELECT_1,
            SpiControllerChipSelect.CHIP_SELECT_2 or SpiControllerChipSelect.CHIP_SELECT_3.

        chipSelectPol : SpiControllerChipSelectPolarity
            Sets the SPI chip select polarity, could be SpiControllerChipSelectPolarity.ACTIVE_LOW or SpiControllerChipSelectPolarity.ACTIVE_HIGH.
        
        frequency : c_uint32
            Sets the SPI Clock frequency in Hz. Currently, the minimum allowed value is 10000 Hz and the maximum allowed value is 50000000 Hz.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.spiControllerInit(id=id, bitOrder=bitOrder, mode=mode, dataWidth=dataWidth, chipSelect=chipSelect, chipSelectPol=chipSelectPol, frequency=frequency)

    @__error_message_parser_wrapper
    def spiControllerSetParameters(self,
                                   id: int,
                                   bitOrder: SpiControllerBitOrder,
                                   mode: SpiControllerMode,
                                   dataWidth: SpiControllerDataWidth,
                                   chipSelect: SpiControllerChipSelect,
                                   chipSelectPol: SpiControllerChipSelectPolarity,
                                   frequency: c_uint32) -> dict:
        """
        This method sets the SPI peripheral configuration.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        bitOrder : SpiControllerBitOrder
            Sets the bit order of the SPI peripheral, could be SpiControllerBitOrder.MSB or SpiControllerBitOrder.LSB.
        
        mode : SpiControllerMode
            Sets the SPI mode, could be SpiControllerMode.MODE_0, SpiControllerMode.MODE_1, SpiControllerMode.MODE_2 or SpiControllerMode.MODE_3.
        
        dataWidth : SpiControllerDataWidth
            Sets the SPI data width, could be SpiControllerDataWidth._8_BITS_DATA, SpiControllerDataWidth._16_BITS_DATA.

        chipSelect : SpiControllerChipSelect
            Sets the SPI chip select, could be SpiControllerChipSelect.CHIP_SELECT_0, SpiControllerChipSelect.CHIP_SELECT_1,
            SpiControllerChipSelect.CHIP_SELECT_2 or SpiControllerChipSelect.CHIP_SELECT_3.

        chipSelectPol : SpiControllerChipSelectPolarity
            Sets the SPI chip select polarity, could be SpiControllerChipSelectPolarity.ACTIVE_LOW or SpiControllerChipSelectPolarity.ACTIVE_HIGH.
        
        frequency : c_uint32
            Sets the SPI Clock frequency in Hz. Currently, the minimum allowed value is 10000 Hz and the maximum allowed value is 50000000 Hz.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.spiControllerSetParameters(id=id, bitOrder=bitOrder, mode=mode, dataWidth=dataWidth, chipSelect=chipSelect, chipSelectPol=chipSelectPol, frequency=frequency)

    @__error_message_parser_wrapper
    def spiControllerTransfer(self, id: int, transferLength: int, payload: list) -> dict:
        """
        This method performs a SPI transfer.
        
        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        transferLength : int
            It is a 2-bytes integer that represents the transfer length. The range allowed is [1, 1024].

        payload : list  
            Python list that contains the SPI data transferred. The list holds bytes elements and
            the maximum length is 1024 bytes.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.spiControllerTransfer(id=id, transferLength=transferLength, payload=payload)

    # GPIO management --------------------------------------------------------------------

    @__error_message_parser_wrapper
    def gpioConfigurePin(self, id: int, pinNumber: GpioPinNumber, functionality: GpioFunctionality, initialOutputLogicLevel: GpioLogicLevel = GpioLogicLevel.LOW) -> dict:
        """
        This method configures a GPIO pin with the specified functionality.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        pinNumber : GpioPinNumber
            An enum representing the GPIO pin number to configure.

        functionality : GpioFunctionality
            An enum representing the desired functionality for the GPIO pin.
        
        initialOutputLogicLevel : GpioLogicLevel
            An enum representing the initial output logic level for the GPIO pin. This parameter is optional
            and only required when the GPIO pin is configured as an output.
            This feature is reserved for future use.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.gpioConfigurePin(id=id, pinNumber=pinNumber, functionality=functionality)

    @__error_message_parser_wrapper
    def gpioDigitalWrite(self, id: int, pinNumber: GpioPinNumber, logicLevel: GpioLogicLevel) -> dict:
        """
        This method writes a digital logic level to a GPIO pin.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        pinNumber : GpioPinNumber
            An enum representing the GPIO pin number to write to.

        logicLevel : GpioLogicLevel
            An enum representing the logic level to write to the GPIO pin.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.gpioDigitalWrite(id=id, pinNumber=pinNumber, logicLevel=logicLevel)

    @__error_message_parser_wrapper    
    def gpioDigitalRead(self, id: int, pinNumber: GpioPinNumber) -> dict:
        """
        This method reads the digital logic level from a GPIO pin.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        pinNumber : GpioPinNumber
            An enum representing the GPIO pin number to read from.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.gpioDigitalRead(id=id, pinNumber=pinNumber)

    @__error_message_parser_wrapper    
    def gpioSetInterrupt(self, id: int, pinNumber: GpioPinNumber, trigger: GpioTriggerType) -> dict:
        """
        This method sets an interruption to a GPIO pin.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        pinNumber : GpioPinNumber
            An enum representing the GPIO pin number to read from.

        trigger : GpioTriggerType
            The trigger type used for the interruption. Must be one of the options provided by the GpioTriggerType enum.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.gpioSetInterrupt(id=id, pinNumber=pinNumber, trigger=trigger) 

    @__error_message_parser_wrapper    
    def gpioDisableInterrupt(self, id: int, pinNumber: GpioPinNumber) -> dict:
        """
        This method disables interruptions of a GPIO pin.

        Parameters
        ----------
        id : int
            It is a 2-bytes integer that represents the transfer id. The range allowed is [1, 65535].

        pinNumber : GpioPinNumber
            An enum representing the GPIO pin number to read from.

        Returns
        -------
        dict
            Return the SystemMessage in dictionary format.

        """
        return self.__driver.gpioDisableInterrupt(id=id, pinNumber=pinNumber)


__all__ = ["supernova"]

from .usb.usb_hid_manager import PULSAR_PID, UsbHidManager

def getConnectedSupernovaDevicesList() -> list:
    '''
    This function can be used to scan all the Supernova devices connected
    to the host computer.

    Arguments
    ---------
    None

    Returns
    -------
    devices: list
        Python list that holds devices dictionary.
    '''

    hwVersionIdentifiers = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
    hwVersionMask = 0xF000

    fwVersionMajorMask = 0x0F00
    fwVersionMinorMask = 0x00F0
    fwVersionPatchMask = 0x000F

    devices = UsbHidManager.enumerate(PULSAR_PID)

    # For each device, convert the release number to a human-readable format.
    # The release number is a 16-bit number where the first 4 bits are the hardware version,
    # the next 4 bits are the major firmware version, the next 4 bits are the minor firmware version,
    # and the last 4 bits are the patch version.
    # The hardware version is represented by a letter from A to H.
    for device in devices:
        device['hardware_version'] = hwVersionIdentifiers[(device['release_number'] & hwVersionMask) >> 12]
        device['firmware_version'] = f'{(device["release_number"] & fwVersionMajorMask) >> 8}.{(device["release_number"] & fwVersionMinorMask) >> 4}.{(device["release_number"] & fwVersionPatchMask)}'

        # Remove unused fields.
        del device['release_number']
        del device['usage_page']
        del device['usage']
        del device['interface_number']
        # When testing, we encountered that this field was not always present.
        # So we need to check if it exists before deleting it.
        if device.get('bus_type'):
            del device['bus_type']

    return devices


from .definitions import *

def i2cControllerInitSerializer(id: c_uint16, busId: I2cBus, frequency_Hz: int, pullUpValue: I2cPullUpResistorsValue) -> tuple[bytes, I2cControllerInitResponse_t]:
    """
    This function performs an I2C_CONTROLLER_INIT command, sending the selected I2C bus
    to initialize and values to set frequency (I2C SCL frequency) and pull up resistors.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : I2cBus
        Selected I2C Bus to initialize.
    frequency_Hz : int
        Numerical value that represents the desired I2C SCL frequency in Hz. Currently, the minimum allowed value
        is 100000 Hz and the maximum allowed value is 1000000 Hz.
    pullUpValue : I2cPullUpResistorsValue
        Value of the pull up enum that represents the desired Pull Up resistors value

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cControllerInitRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_CONTROLLER_INIT.value
    command.fields.parameters.busId = busId.value
    command.fields.parameters.frequency_Hz = frequency_Hz
    command.fields.parameters.pullUpValue = pullUpValue.value

    # Return command
    return command.toBytes(), I2cControllerInitResponse_t()

def i2cControllerSetParametersSerializer(id: c_uint16, busId: I2cBus, frequency_Hz: int, pullUpValue: I2cPullUpResistorsValue) -> tuple[bytes, I2cControllerSetParametersResponse_t]:
    """
    This function performs an I2C_SET_PARAMETERS command, sending the selected I2C bus
    to configure and values to set frequency (I2C SCL frequency) and pull up resistors.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : I2cBus
        Selected I2C Bus to set parameters.
    frequency_Hz : int
        Numerical value that represents the desired I2C SCL frequency in Hz. Currently, the minimum allowed value
            is 100000 Hz and the maximum allowed value is 1000000 Hz.
    pullUpValue : I2cPullUpResistorsValue
        Value of the pull up enum that represents the desired Pull Up resistors value.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cControllerSetParametersRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_CONTROLLER_SET_PARAMETERS.value
    command.fields.parameters.busId = busId.value
    command.fields.parameters.frequency_Hz = frequency_Hz
    command.fields.parameters.pullUpValue = pullUpValue.value

    # Return command
    return command.toBytes(), I2cControllerSetParametersResponse_t()

def i2cSetPullUpResistorsSerializer(id: c_uint16, busId: I2cBus, pullUpValue: I2cPullUpResistorsValue) -> tuple[bytes, I2cSetPullUpResistorsResponse_t]:
    """
    This function performs an I2C_SET_PULL_UP_RESISTORS command, sending the selected pull up
    resistor value for the I2C lines.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : I2cBus
        Selected I2C Bus to set pull up resistors.
    pullUpValue : I2cPullUpResistorsValue
        This parameter represents the different values for the pull up resistors.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cSetPullUpResistorsRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_SET_PULLUP_RESISTORS.value
    command.fields.parameters.busId = busId.value
    command.fields.parameters.pullUpValue = pullUpValue.value

    # Return command
    return command.toBytes(), I2cSetPullUpResistorsResponse_t()

def i2cControllerWriteSerializer(id: c_uint16, busId: I2cBus, targetAddress: int, registerAddress: list, payload: list, isNonStop: bool, is10BitTargetAddress: bool) -> tuple[bytes, I2cControllerTransferResponse_t]:
    """
    This function is used to generate the I2C_CONTROLLER_WRITE command.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : int
        Selected I2C Bus to write data.
    targetAddress : int
        7-bit slave address
    registerAddress : list
        List that represents the subaddress/register address. It can contains
        up to 4 bytes.
    payload : list
        List containing bytes to transfer.
    isNonStop : bool
        This parameter is used to indicate if the transaction is non-stop.
    is10BitTargetAddress : bool
        This parameter is used to indicate if the target address is 10-bit.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cControllerTransferRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_CONTROLLER_WRITE.value

    # I2C Bus
    command.fields.parameters.busId = busId.value

    # Transfer Length
    command.fields.parameters.transfLength = len(payload)

    # 10-bit address
    if is10BitTargetAddress:
        # Two MSB of the target address
        command.fields.parameters.targetAddress = I2C_10_BIT_ADDRESS_FIRST_BYTE_MASK | (targetAddress >> I2C_10_BIT_ADDRESS_FIRST_BYTE_SHIFT)

        # 8 LSB of the target address
        command.fields.parameters.registerAddress = c_uint32(targetAddress & I2C_10_BIT_ADDRESS_SECOND_BYTE_MASK)
        command.fields.parameters.registerAddressLength = 1

    else:
        # 7-bit address
        command.fields.parameters.targetAddress = targetAddress

    # Register address length
    registerAddressLength = len(registerAddress)
    command.fields.parameters.registerAddressLength += registerAddressLength

    # Register address
    for i in range(registerAddressLength):
        command.fields.parameters.registerAddress = c_uint32(command.fields.parameters.registerAddress << 8 | registerAddress[i])

    # Payload
    for i in range(len(payload)):
        command.fields.payload[i] = payload[i]
    
    # Non-stop
    command.fields.parameters.isNonStop = isNonStop

    # Return command structure.
    return command.toBytes(), I2cControllerTransferResponse_t()

def i2cControllerReadSerializer(id: c_uint16, busId: I2cBus, targetAddress: int, length: int, registerAddress: list, is10BitTargetAddress: bool) -> tuple[bytes, I2cControllerTransferResponse_t]:
    """
    This function is used to generate the I2C_CONTROLLER_READ command.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : I2cBus
        Selected I2C Bus to read data.
    targetAddress : int
        7-bit slave address
    length : int
        Length of data to be read from the USB device.
    registerAddress : list
        List that represents the subaddress/register address. It can contains
        up to 4 bytes.
    is10BitTargetAddress : bool
        This parameter is used to indicate if the target address is 10-bit.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cControllerTransferRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_CONTROLLER_READ.value

    # I2C Bus
    command.fields.parameters.busId = busId.value

    # Transfer Length
    command.fields.parameters.transfLength = length

    # 10-bit address
    if is10BitTargetAddress:
        # Two MSB bits of the target address
        command.fields.parameters.targetAddress = I2C_10_BIT_ADDRESS_FIRST_BYTE_MASK | (targetAddress >> I2C_10_BIT_ADDRESS_FIRST_BYTE_SHIFT)

        # 8-bit LSB of the target address
        command.fields.parameters.registerAddress = c_uint32(targetAddress & I2C_10_BIT_ADDRESS_SECOND_BYTE_MASK)
        command.fields.parameters.registerAddressLength = 1

    else:
        # 7-bit address
        command.fields.parameters.targetAddress = targetAddress

    # Register address length
    registerAddressLength = len(registerAddress)
    command.fields.parameters.registerAddressLength += registerAddressLength

    # Register address
    for i in range(registerAddressLength):
        command.fields.parameters.registerAddress = c_uint32(command.fields.parameters.registerAddress << 8 | registerAddress[i])

    # Return command structure.
    return command.toBytes(), I2cControllerTransferResponse_t()

def i2cControllerScanBusSerializer(id: c_uint16, busId: I2cBus, include10BitAddresses: bool) -> tuple[bytes, I2cControllerScanBusResponse_t]:
    """
    This function is used to generate the I2C_CONTROLLER_SCAN_BUS command.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    busId : I2cBus
        Selected I2C Bus to scan.
    include10BitAddresses : bool
        This parameter is used to indicate if the scan should include 10-bit addresses.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = I2cControllerScanBusRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = I2cCommandCodes.I2C_CONTROLLER_SCAN_BUS.value
    command.fields.parameters.busId = busId.value
    command.fields.parameters.include10BitAddresses = include10BitAddresses

    # Return command structure.
    return command.toBytes(), I2cControllerScanBusResponse_t()



__all__ = ["Pulsar"]

from .supernovasdk.binhosupernova import getConnectedSupernovaDevicesList

def getConnectedPulsarDevicesList() -> list:
    '''
    This function can be used to scan all the Pulsar devices connected
    to the host computer.

    Arguments
    ---------
    None

    Returns
    -------
    devices: list
        Python list that holds devices dictionary.
    '''
    # Return list of devices.
    return getConnectedSupernovaDevicesList()



from .serializers import *
from ..helpers.validator import check_type, check_valid_id, check_range
from ...utils.system_message import SystemMessage, SystemModules, RequestValidatorOpcode

#===============================================================================
# region Validators
#===============================================================================

def resetDeviceValidator(metadata: dict):
    """
    This function validates the metadata for the RESET DEVICE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "RESET DEVICE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = resetDeviceSerializer(metadata["id"])

    return request, response, result

def enterBootModeValidator(metadata: dict):
    """
    This function validates the metadata for the ENTER BOOT MODE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "ENTER BOOT MODE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = enterBootModeSerializer(metadata["id"])

    return request, response, result

def enterIspModeValidator(metadata: dict):
    """
    This function validates the metadata for the ENTER ISP MODE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "ENTER ISP MODE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = enterIspModeSerializer(metadata["id"])

    return request, response, result

def getUsbStringValidator(metadata: dict):
    """
    This function validates the metadata for the GET USB STRING command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "GET USB STRING request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    if (not check_type(metadata["subcommand"], GetUsbStringSubCommand)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for subcommand value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = getUsbStringSerializer(metadata["id"], metadata["subcommand"])

    return request, response, result

def setI3cVoltageValidator(metadata: dict):
    """
    This function validates the metadata for the SET I3C VOLTAGE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "SET I3C VOLTAGE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    if (metadata["voltage_mV"]!=POWER_OFF_VOLTAGE) and (not check_range(metadata["voltage_mV"], int, MIN_I3C_VOLTAGE_VALUE, MAX_I3C_VOLTAGE_VALUE)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: voltage value out of range"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = setI3cVoltageSerializer(metadata["id"], metadata["voltage_mV"])

    return request, response, result

def setI2cSpiUartGpioVoltValidator(metadata: dict):
    """
    This function validates the metadata for the SET I2C SPI UART GPIO VOLTAGE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "SET I2C SPI UART GPIO VOLTAGE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    if (metadata["voltage_mV"]!=POWER_OFF_VOLTAGE) and (not check_range(metadata["voltage_mV"], int, MIN_I2C_SPI_UART_GPIO_VOLTAGE_VALUE, MAX_I2C_SPI_UART_GPIO_VOLTAGE_VALUE)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: voltage value out of range"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = setI2cSpiUartGpioVoltageSerializer(metadata["id"], metadata["voltage_mV"])

    return request, response, result

def useExternalI3cVoltageValidator(metadata: dict):
    """
    This function validates the metadata for the USE EXT SRC I3C VOLTAGE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "USE EXT SRC I3C VOLTAGE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = useExternalI3cVoltageSerializer(metadata["id"])

    return request, response, result

def useExternalI2cSpiUartGpioVoltageValidator(metadata: dict):
    """
    This function validates the metadata for the USE EXT SRC I2C-SPI-UART-GPIO VOLTAGE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "USE EXT SRC I2C-SPI-UART-GPIO VOLTAGE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = useExternalI2cSpiUartGpioVoltageSerializer(metadata["id"])

    return request, response, result

def getAnalogMeasurementsValidator(metadata: dict):
    """
    This function validates the metadata for the GET ANALOG MEASUREMENTS command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "GET ANALOG MEASUREMENTS request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = getAnalogMeasurementsSerializer(metadata["id"])

    return request, response, result

def getI3cConnectorsStatusValidator(metadata: dict):
    """
    This function validates the metadata for the GET I3C CONNECTORS STATUS command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "GET I3C CONNECTORS STATUS request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = getI3cConnectorsStatusSerializer(metadata["id"])

    return request, response, result

def getDeviceInfoValidator(metadata: dict):
    """
    This function validates the metadata for the GET DEVICE INFO command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "GET DEVICE INFO request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = getDeviceInfoSerializer(metadata["id"])

    return request, response, result

# endregion



from ..common_definitions import *

#================================================================================#
# region UART COMMAND DEFINITIONS
#================================================================================#

class UartCommandCodes(Enum):
    """
    Enumerator of the UART command codes.
    """
    UART_DEINIT                 = makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 1)
    UART_INIT                   = makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 2)
    UART_SET_PARAMETERS         = makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 3)
    UART_GET_PARAMETERS         = makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 4)
    UART_SEND                   = makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 5)
    UART_RECEIVE_NOTIFICATION	= makeCommandCode(Group.UART.value, CommandRole.GENERIC.value, CommandType.NOTIFICATION.value, 1)

UART_COMMAND_NAMES = {
    result.value: result.name.replace("_", " ") for result in UartCommandCodes
}

# endregion

#================================================================================#
# region UART COMMON DEFINITIONS
#================================================================================#

# UART Bus Configuration
class UartBaudRate(Enum):
    """
    This enum represents the UART baudrate options.
    """
    UART_BAUD_600    = 0x00
    UART_BAUD_1200   = 0x01
    UART_BAUD_2400   = 0x02
    UART_BAUD_4800   = 0x03
    UART_BAUD_9600   = 0x04
    UART_BAUD_14400  = 0x05
    UART_BAUD_19200  = 0x06
    UART_BAUD_38400  = 0x07
    UART_BAUD_56000  = 0x08
    UART_BAUD_57600  = 0x09
    UART_BAUD_115200 = 0x0A

class UartParity(Enum):
    """ 
    This enum represents the UART parity options.
    """
    UART_NO_PARITY   = 0x00
    UART_EVEN_PARITY = 0x01
    UART_ODD_PARITY  = 0x02

class UartDataSize(Enum):
    """
    This enum represents the UART data character size options.
    """
    UART_7BIT_BYTE = 0x00
    UART_8BIT_BYTE = 0x01

class UartStopBit(Enum):
    """
    This enum represents the UART stop bit options.
    """
    UART_ONE_STOP_BIT = 0x00
    UART_TWO_STOP_BIT = 0x01

class UartConfigurationParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("baudRate", c_uint8),
                ("hardwareHandshake", c_uint8),
                ("parityMode", c_uint8),
                ("dataSize", c_uint8),
                ("stopBitType", c_uint8)]

# endregion

#================================================================================#
# region UART INIT
#================================================================================#

# Request ---------------------------------------------------------------------- #

class UartInitRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", UartConfigurationParameters_t)]

UartInitRequestArray_t = c_uint8 * sizeof(UartInitRequestFields_t)

class UartInitRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", UartInitRequestArray_t),
                ("fields", UartInitRequestFields_t)]

# Response --------------------------------------------------------------------- #

class UartInitResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

UartInitResponseArray_t = c_uint8 * sizeof(UartInitResponseFields_t)

class UartInitResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", UartInitResponseArray_t ),
                ("fields", UartInitResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = UartInitResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": UART_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region UART SET PARAMETERS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class UartSetParametersRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", UartConfigurationParameters_t)]

UartSetParametersRequestArray_t = c_uint8 * sizeof(UartSetParametersRequestFields_t)

class UartSetParametersRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", UartSetParametersRequestArray_t),
                ("fields", UartSetParametersRequestFields_t)]

# Response --------------------------------------------------------------------- #

class UartSetParametersResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

UartSetParametersResponseArray_t = c_uint8 * sizeof(UartSetParametersResponseFields_t)

class UartSetParametersResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", UartSetParametersResponseArray_t ),
                ("fields", UartSetParametersResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = UartSetParametersResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": UART_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region UART SEND
#================================================================================#

MAX_UART_TRANSFER_LENGTH = 1024

# Request ---------------------------------------------------------------------- #

class UartSendRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("payloadLength", c_uint16)]

UartSendRequestPayload_t = c_uint8 * MAX_UART_TRANSFER_LENGTH

class UartSendRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", UartSendRequestParameters_t),
                ("payload", UartSendRequestPayload_t)]

UartSendRequestArray_t = c_uint8 * sizeof(UartSendRequestFields_t)

class UartSendRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", UartSendRequestArray_t),
                ("fields", UartSendRequestFields_t)]

    def toBytes(self) -> bytes:
        """
        Method to return an bytes object representing the command
        serialization.
        """
        length = sizeof(CommandRequestHeader_t) + sizeof(UartSendRequestParameters_t) + self.fields.parameters.payloadLength
        return bytes(self.data[:length])

# Response --------------------------------------------------------------------- #

class UartSendResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

UartSendResponseArray_t = c_uint8 * sizeof(UartSendResponseFields_t)

class UartSendResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", UartSendResponseArray_t),
                ("fields", UartSendResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = UartSendResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": UART_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region UART RECEIVE NOTIFICATION
#================================================================================#

class UartReceiveNotificationParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("payloadLength", c_uint16)]

UartReceiveNotificationPayload_t = c_uint8 * MAX_UART_TRANSFER_LENGTH

class UartReceiveNotificationFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", UartReceiveNotificationParameters_t),
                ("payload", UartReceiveNotificationPayload_t)]

UartReceiveNotificationArray_t = c_uint8 * sizeof(UartReceiveNotificationFields_t)

class UartReceiveNotification_t(Union):
    _fields_ = [("data", UartReceiveNotificationArray_t),
                ("fields", UartReceiveNotificationFields_t)]

    def fromBytes(self, data):
        for i in range(len(data)):
            self.data[i] = data[i]

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": UART_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result],
            "payload_length": self.fields.parameters.payloadLength,
            "payload": self.fields.payload[:self.fields.parameters.payloadLength]
        }

    def __str__(self) -> str:
        return str(self.toDictionary())

# endregion


from .serializers import *
from ..helpers.validator import check_type, check_range, check_valid_id, check_byte_array, getRepeatedItems
from ...utils.system_message import SystemMessage, SystemModules, RequestValidatorOpcode

#================================================================================#
# region Helper functions
#================================================================================#

def validateI3cControllerConfiguration(metadata: dict, result: SystemMessage):
    """
    This function validates the metadata for the commands that configure the I3C controller.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.
    result : SystemMessage
        SystemMessage to be updated.

    """

    notValidFrequenciesPair = [ (I3cPushPullTransferRate.PUSH_PULL_2_5_MHZ_10_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_2_5_MHZ_15_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_2_5_MHZ_20_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_2_5_MHZ_25_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_2_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_2_5_MHZ_25_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_3_125_MHZ_12_5_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_3_125_MHZ_18_75_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_3_125_MHZ_25_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_3_125_MHZ_31_25_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_2_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_3_125_MHZ_31_25_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_5_MHZ_40_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_5_MHZ_50_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_2_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_5_MHZ_50_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ),
                                (I3cPushPullTransferRate.PUSH_PULL_6_25_MHZ_50_DC, I3cOpenDrainTransferRate.OPEN_DRAIN_4_17_MHZ) ]

    if (not check_type(metadata["pushPullRate"], I3cPushPullTransferRate)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for pushPullRate value"
    if (not check_type(metadata["i3cOpenDrainRate"], I3cOpenDrainTransferRate)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for i3cOpenDrainRate value"
    if (not check_type(metadata["i2cOpenDrainRate"], I2cTransferRate)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for i2cOpenDrainRate value"
    if (not check_type(metadata["driveStrength"], I3cDriveStrength)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for driveStrength value"
    if (metadata["pushPullRate"], metadata["i3cOpenDrainRate"]) in notValidFrequenciesPair:
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: invalid frequency pair"

def validateI3cControllerTransfer(metadata: dict, result: SystemMessage, isWrite: bool):
    """
    This function validates the metadata for the commands that perform I3C controller transfers.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.
    result : SystemMessage
        SystemMessage to be updated.
    isWrite : bool
        Flag to indicate if the transfer is a write operation.

    """
    if (not check_type(metadata["mode"], TransferMode)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for mode value"
    if (not check_type(metadata["registerAddress"], list)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for registerAddress value"
    if (not check_byte_array(metadata["registerAddress"],MAX_REGISTER_ADDRESS_LENGTH)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: registerAddress array length out of range or wrong type"
    if isWrite:
        if (not check_type(metadata["data"], list)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: wrong type for data value"
        if (not check_byte_array(metadata["data"],MAX_I3C_PRIVATE_TRANSFER_LENGTH)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: data array length out of range or wrong type"
    else:
        if (not check_range(metadata["length"], int, 1, MAX_I3C_PRIVATE_TRANSFER_LENGTH)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: length value out of range"

def validateI3cTargetConfiguration(metadata: dict, result: SystemMessage):
    """
    This function validates the metadata for the commands that configure the I3C target.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.
    result : SystemMessage
        SystemMessage to be updated.

    """
    if (not check_type(metadata["memoryLayout"], I3cTargetMemoryLayout_t)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for memory layout value"
    if (not check_type(metadata["pid"], list)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for PID value"
    if (not check_byte_array(metadata["pid"], sizeof(I3cPidBytes_t))):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: PID length should be 6 bytes"
    if (not check_range(metadata["bcr"], int, MIN_I3C_BCR_VALUE, MAX_I3C_BCR_VALUE)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = f"ARGUMENT ERROR: BCR out of range [{MIN_I3C_BCR_VALUE},{MAX_I3C_BCR_VALUE}]"
    if (not check_type(metadata["dcr"], I3cTargetDcr_t)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for DCR value"
    if (not check_range(metadata["staticAddress"], int, 0, 127)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: Static Address out of range [0,127]"

    i2c_reserved_addresses = list(range(0x00, 0x08)) + list(range(0x78, 0x80))

    if (metadata["staticAddress"] in i2c_reserved_addresses):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: Invalid Static Address value. Reserved by I2C protocol"

    if (not check_range(metadata["mwl"], int, 0, I3C_TARGET_MAX_MWL)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = f"ARGUMENT ERROR: Maximum Write Length out of range [0,{I3C_TARGET_MAX_MWL}]"
    if (not check_range(metadata["mrl"], int, 0, I3C_TARGET_MAX_MRL)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = f"ARGUMENT ERROR: Maximum Read Length out of range [0,{I3C_TARGET_MAX_MRL}]"

def validateI3cTargetTransfer(metadata: dict, result: SystemMessage, isWrite: bool):
    """
    This function validates the metadata for the commands that perform I3C Target transfers.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.
    result : SystemMessage
        SystemMessage to be updated.
    isWrite : bool
        Flag to indicate if the transfer is a write operation.

    """
    if (not check_range(metadata["memoryAddress"],int,0,I3C_TARGET_MEMORY_SIZE)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: Invalid memory address value"
    if isWrite:
        if (not check_type(metadata["data"], list)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: wrong type for data value"
        if (len(metadata["data"]) == 0):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: no data input"
        if (not check_byte_array(metadata["data"],I3C_TARGET_MEMORY_SIZE)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: invalid data input"
    else:
        if (not check_range(metadata["length"],int,0,I3C_TARGET_MEMORY_SIZE)):
            result.opcode = RequestValidatorOpcode.FAIL
            result.message = "ARGUMENT ERROR: invalid length value"

# endregion

#================================================================================#
# region I3C CONTROLLER INIT validator
#================================================================================#

def i3cControllerInitValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER INIT command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER INIT request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    
    validateI3cControllerConfiguration(metadata, result)
    

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerInitSerializer(metadata["id"], metadata["pushPullRate"], metadata["i3cOpenDrainRate"], metadata["i2cOpenDrainRate"], metadata["driveStrength"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER SET PARAMETERS validator
#================================================================================#

def i3cControllerSetParametersValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER SET PARAMETERS command to perform an I3C Private Read Transfer.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER SET PARAMETERS request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    
    validateI3cControllerConfiguration(metadata, result)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerSetParametersSerializer(metadata["id"], metadata["pushPullRate"], metadata["i3cOpenDrainRate"], metadata["i2cOpenDrainRate"], metadata["driveStrength"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER INIT BUS validator
#================================================================================#

def i3cControllerInitBusValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER INIT BUS command.
    Validates the targetDeviceTable of the user request. Checks that all I3C devices have a DAA method assigned and that there
    are no repeated addresses:
    - static address for I2C targets and I3C targets to be initialized with SETAASA
    - static and dynamic addresses for I3C targets to be initialized with SETDASA
    - dynamic addresses for I3C targets to be initialized with ENTDAA

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """

    request = None
    response = None

    def verifyI3cDynamicAddressAssignmentMethod(targetDeviceTable: dict):
        """
        Verifies that all I3C entries from the targetDeviceTable have a DAA method assigned.

        Arguments
        ---------
        dict: dictionary of metadata that represents the targetDeviceTable for I3cInitBusRequest_t

        Returns
        -------
        indexOfAddrWithoutMethod
            Index of the entries from targetDeviceTable that does not indicate a DAA method.

        """

        indexOfAddrWithoutMethod = []
        for index, target in targetDeviceTable.items():
            if ((target["configuration"]["targetType"] == TargetType.I3C_DEVICE) and
                ( target["configuration"]["daaUseSETDASA"] == False) and ( target["configuration"]["daaUseSETAASA"] == False) and ( target["configuration"]["daaUseENTDAA"] == False)):
                indexOfAddrWithoutMethod.append(index)
        return indexOfAddrWithoutMethod

    def getRepeatedAddresses(targetDeviceTable: dict):
        """
        Gets all the repeated addresses from the targetDeviceTable argument.

        Arguments
        ---------
        dict: dictionary of metadata that represents the targetDeviceTable for I3cInitBusRequest_t

        Returns
        -------
        listOfAddresses
           List of Addresses repeated in targetDeviceTable.

        """

        # Address used for SETDASA point to point
        SETDASA_POINT_TO_POINT_ADDR = 0x01

        listOfAddresses = []

        for target in targetDeviceTable.values():
            if (target["configuration"]["targetType"] == TargetType.I3C_DEVICE):

                # If the I3C device supports SETDASA and its static and dynamic addresses are SETDASA_POINT_TO_POINT_ADDR it might refer to a point-to-point SETDASA
                if not((target["configuration"]["daaUseSETDASA"] == True) and (target['staticAddress'] == SETDASA_POINT_TO_POINT_ADDR) and (target['dynamicAddress'] == SETDASA_POINT_TO_POINT_ADDR)):

                    if (target["configuration"]["daaUseSETDASA"] == True) or (target["configuration"]["daaUseENTDAA"] == True):
                        listOfAddresses.append(target['dynamicAddress'])

                    if (target["configuration"]["daaUseSETDASA"] == True) or (target["configuration"]["daaUseSETAASA"] == True):
                        listOfAddresses.append(target['staticAddress'])

            if (target["configuration"]["targetType"] == TargetType.I2C_DEVICE):
                listOfAddresses.append(target['staticAddress'])

        # Return the list of repeated addresses
        return getRepeatedItems(listOfAddresses)

    if metadata.get("targetDeviceTable") is not None:
        targetDeviceTable = metadata["targetDeviceTable"]

        listOfTargetsWithoutDaaMethod = verifyI3cDynamicAddressAssignmentMethod(targetDeviceTable)
        if listOfTargetsWithoutDaaMethod:
            targets_str = ', '.join([f"{target_index}" for target_index in listOfTargetsWithoutDaaMethod])
            message = f"I3C CONTROLLER INIT BUS failed: target/s in position {targets_str} of the input table not supporting any DAA method"
            result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.FAIL, message)
            return request, response, result

        listOfRepeatedAddr = getRepeatedAddresses(targetDeviceTable)
        if listOfRepeatedAddr:
            addresses_str = ', '.join([f"0x{addr:02X}" for addr in listOfRepeatedAddr])
            message = f"I3C CONTROLLER INIT BUS failed: address/es {addresses_str} repeated"
            result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.FAIL, message)
            return request, response, result

    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER INIT BUS request success")

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerInitBusSerializer(metadata["id"], metadata["targetDeviceTable"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER RESET BUS validator
#================================================================================#

def i3cControllerResetBusValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER RESET BUS command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER RESET BUS request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerResetBusSerializer(metadata["id"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER GET TARGET DEVICES TABLE validator
#================================================================================#

def i3cControllerGetTargetDevicesTableValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER GET TARGET DEVICES TABLE command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER GET TARGET DEVICES TABLE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerGetTargetDevicesTableSerializer(metadata["id"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER SET TARGET DEVICE CONFIGURATION validator
#================================================================================#

def i3cControllerSetTargetDeviceConfigurationValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER SET TARGET DEVICE CONFIGURATION command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C SET TARGET DEVICE CONFIG request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerSetTargetDeviceConfigurationSerializer(metadata["id"], metadata["targetAddress"], metadata["configuration"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER PRIVATE WRITE validator
#================================================================================#

def i3cControllerWriteValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER PRIVATE TRANSFER command to perform an I3C Private Write Transfer.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER PRIVATE WRITE request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    validateI3cControllerTransfer(metadata, result, isWrite=True)
    
    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerWriteSerializer(metadata["id"], metadata["targetAddress"], metadata["mode"], metadata["registerAddress"], metadata["data"], metadata["startWith7E"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER PRIVATE READ validator
#================================================================================#

def i3cControllerReadValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER PRIVATE TRANSFER command to perform an I3C Private Read Transfer.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER PRIVATE READ request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    
    validateI3cControllerTransfer(metadata, result, isWrite=False)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerReadSerializer(metadata["id"], metadata["targetAddress"], metadata["mode"], metadata["registerAddress"], metadata["length"], metadata["startWith7E"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER CCC TRANSFER validator
#================================================================================#

def i3cControllerCccTransferValidator(metadata: dict):
    """
    This function validates the metadata for the I3C CONTROLLER CCC TRANSFER command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C SEND CCC request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    if (not check_type(metadata["mode"], TransferMode)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for mode value"
    if (not check_type(metadata["direction"], TransferDirection)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for direction value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerCccTransferSerializer(metadata["id"], metadata["targetAddress"], metadata["direction"], metadata["mode"], metadata["commandType"], metadata["defByte"], metadata["ccc"], metadata["length"], metadata["data"])

    return request, response, result

#endregion

#================================================================================#
# region I3C CONTROLLER TRIGGER PATTERN validator
#================================================================================#

def i3cControllerTriggerPatternValidator(metadata: dict):
    """
    This function validates the metadata for the I3C TRIGGER PATTERN command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C CONTROLLER TRIGGER PATTERN request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    if (not check_type(metadata["pattern"], I3cPattern)):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong type for pattern value"

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cControllerTriggerPatternSerializer(metadata["id"],  metadata["pattern"])

    return request, response, result

#endregion

#================================================================================#
# region I3C TARGET validators
#================================================================================#

def i3cTargetInitValidator(metadata: dict):
    """
    This function validates the metadata for the I3C TARGET INIT command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C TARGET INIT request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"
    
    validateI3cTargetConfiguration(metadata, result)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cTargetInitSerializer(metadata["id"], metadata["memoryLayout"], metadata["pid"], metadata["bcr"], metadata["dcr"], metadata["staticAddress"], metadata["mwl"], metadata["mrl"])

    return request, response, result

def i3cTargetSetParametersValidator(metadata: dict):
    """
    This function validates the metadata for the I3C TARGET SET PARAMETERS command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C TARGET SET PARAMETERS request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "ARGUMENT ERROR: wrong id value"

    validateI3cTargetConfiguration(metadata, result)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cTargetSetParametersSerializer(metadata["id"], metadata["memoryLayout"], metadata["pid"], metadata["bcr"], metadata["dcr"], metadata["staticAddress"], metadata["mwl"], metadata["mrl"])

    return request, response, result

def i3cTargetWriteMemoryValidator(metadata: dict):
    """
    This function validates the metadata for the I3C TARGET WRITE MEMORY command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C TARGET WRITE MEMORY request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "I3C TARGET WRITE MEMORY request failed, wrong id value"
    
    validateI3cTargetTransfer(metadata, result, isWrite=True)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cTargetWriteMemorySerializer(metadata["id"], metadata["memoryAddress"], metadata["data"])

    return request, response, result

def i3cTargetReadMemoryValidator(metadata: dict):
    """
    This function validates the metadata for the I3C TARGET READ MEMORY command.

    Arguments
    ---------
    metadata : dict
        Metadata to be validated.

    Returns
    -------
    tuple
        A tuple containing the request, response and result of the validation.

    """
    result = SystemMessage(SystemModules.VALIDATION, RequestValidatorOpcode.SUCCESS, "I3C TARGET READ MEMORY request success")
    request = None
    response = None

    if (not check_valid_id(metadata["id"])):
        result.opcode = RequestValidatorOpcode.FAIL
        result.message = "I3C TARGET READ MEMORY request failed, wrong id value"

    validateI3cTargetTransfer(metadata, result, isWrite=False)

    if (result.opcode == RequestValidatorOpcode.SUCCESS):
        request, response = i3cTargetReadMemorySerializer(metadata["id"], metadata["memoryAddress"], metadata["length"])

    return request, response, result

#endregion


from ..common_definitions import *

#================================================================================#
# region I2C COMMAND DEFINITIONS
#================================================================================#

class I2cCommandCodes(Enum):
    """
    Enumeration of I2C command codes.
    """
    I2C_DEINIT                      = makeCommandCode(Group.I2C.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 1)
    I2C_SET_PULLUP_RESISTORS        = makeCommandCode(Group.I2C.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 2)
    I2C_CONTROLLER_INIT             = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 1)
    I2C_CONTROLLER_SET_PARAMETERS   = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 2)
    I2C_CONTROLLER_GET_PARAMETERS   = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 3)
    I2C_CONTROLLER_WRITE            = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 4)
    I2C_CONTROLLER_READ             = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 5)
    I2C_CONTROLLER_SCAN_BUS         = makeCommandCode(Group.I2C.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 6)

I2C_COMMAND_NAMES = {
    result.value: result.name.replace("_", " ") for result in I2cCommandCodes
}

# endregion

#================================================================================#
# region I2C COMMON DEFINITIONS
#================================================================================#

# I2C Bus Configuration
I2C_CONTROLLER_MIN_FREQUENCY = 100000
I2C_CONTROLLER_MAX_FREQUENCY = 1000000

class I2cPullUpResistorsValue(Enum):
    """
    This enum represents the total value of the I2C Pull-Up resistors
    """
    I2C_PULLUP_150Ohm   = 0x00
    I2C_PULLUP_220Ohm   = 0x01
    I2C_PULLUP_330Ohm   = 0x02
    I2C_PULLUP_470Ohm   = 0x03
    I2C_PULLUP_680Ohm   = 0x04
    I2C_PULLUP_1kOhm    = 0x05
    I2C_PULLUP_1_5kOhm  = 0x06
    I2C_PULLUP_2_2kOhm  = 0x07
    I2C_PULLUP_4kOhm    = 0x08
    I2C_PULLUP_DISABLE  = 0x09

class I2cBus(Enum):
    """
    This enum represent the I2C selected bus
    """
    I2C_BUS_A                       = 0x00
    I2C_BUS_B                       = 0x01

# Definitions for I2C address validation
I2C_MIN_ADDRESS = 0x00
I2C_MAX_10_BIT_ADDRESS = 0x3FF

# Definitions for 10-bit I2C address
I2C_10_BIT_ADDRESS_FIRST_BYTE_MASK  = 0x78
I2C_10_BIT_ADDRESS_FIRST_BYTE_SHIFT = 8
I2C_10_BIT_ADDRESS_SECOND_BYTE_MASK = 0xFF

# Definitions for I2C register address
I2C_7_BIT_REGISTER_ADDRESS_MAX_LENGTH = 4
I2C_10_BIT_REGISTER_ADDRESS_MAX_LENGTH = 3

# endregion

#================================================================================#
# region I2C RESULT DEFINITIONS
#================================================================================#

class I2cResultCodes(Enum):
    """
    Enumeration of I2C result codes.
    """
    I2C_PULLUP_RESISTOR_SETTING_FAILURE = makeResultCode(Group.I2C.value, 1)
    I2C_ARBITRATION_LOST				= makeResultCode(Group.I2C.value, 2)
    I2C_NACK_ADDRESS					= makeResultCode(Group.I2C.value, 3)
    I2C_NACK_BYTE						= makeResultCode(Group.I2C.value, 4)
    I2C_BIT_ERROR						= makeResultCode(Group.I2C.value, 5)
    I2C_START_STOP_ERROR				= makeResultCode(Group.I2C.value, 6)
    I2C_BUSY							= makeResultCode(Group.I2C.value, 7)
    I2C_TIMEOUT_CONTINUE_TRANSFER		= makeResultCode(Group.I2C.value, 8)
    I2C_TIMEOUT_WAITING_BUS_EVENT		= makeResultCode(Group.I2C.value, 9)
    I2C_TIMEOUT_SCL_LOW					= makeResultCode(Group.I2C.value, 10)
    I2C_DMA_REQUEST_FAIL				= makeResultCode(Group.I2C.value, 11)
    I2C_INVALID_PARAMETER				= makeResultCode(Group.I2C.value, 12) 
    I2C_NO_TRANSFER_IN_PROGRESS			= makeResultCode(Group.I2C.value, 13)
    I2C_UNEXPECTED_STATE				= makeResultCode(Group.I2C.value, 14)
    I2C_BUS_WITH_NO_TARGETS_CONNECTED   = makeResultCode(Group.I2C.value, 15)

I2C_RESULT_NAMES = {
    **COMMON_RESULT_NAMES,
    **{result.value: result.name for result in I2cResultCodes}
}

# endregion

#================================================================================#
# region I2C CONTROLLER INIT
#================================================================================#

# Request ---------------------------------------------------------------------- #

class I2cControllerInitRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
                ("frequency_Hz", c_uint32),
                ("pullUpValue", c_uint8)]

class I2cControllerInitRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", I2cControllerInitRequestParameters_t)]

I2cControllerInitRequestArray_t = c_uint8 * sizeof(I2cControllerInitRequestFields_t)

class I2cControllerInitRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", I2cControllerInitRequestArray_t ),
                ("fields", I2cControllerInitRequestFields_t )]

# Response --------------------------------------------------------------------- #

class I2cControllerInitResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8)]

class I2cControllerInitResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", I2cControllerInitResponseParameters_t)]

I2cControllerInitResponseArray_t = c_uint8 * sizeof(I2cControllerInitResponseFields_t)

class I2cControllerInitResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", I2cControllerInitResponseArray_t ),
                ("fields", I2cControllerInitResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = I2cControllerInitResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": I2C_COMMAND_NAMES[self.fields.header.code],
            "result": I2C_RESULT_NAMES[self.fields.header.result],
            "i2c_bus": I2cBus(self.fields.parameters.busId).name
        }

# endregion

#================================================================================#
# region I2C CONTROLLER SET PARAMETERS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class I2cControllerSetParametersRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
                ("frequency_Hz", c_uint32),
                ("pullUpValue", c_uint8)]

class I2cControllerSetParametersRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", I2cControllerSetParametersRequestParameters_t)]

I2cControllerSetParametersRequestArray_t = c_uint8 * sizeof(I2cControllerSetParametersRequestFields_t)

class I2cControllerSetParametersRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", I2cControllerSetParametersRequestArray_t ),
                ("fields", I2cControllerSetParametersRequestFields_t )]

# Response --------------------------------------------------------------------- #

class I2cControllerSetParametersResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8)]

class I2cControllerSetParametersResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", I2cControllerSetParametersResponseParameters_t)]

I2cControllerSetParametersResponseArray_t = c_uint8 * sizeof(I2cControllerSetParametersResponseFields_t)

class I2cControllerSetParametersResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", I2cControllerSetParametersResponseArray_t ),
                ("fields", I2cControllerSetParametersResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = I2cControllerSetParametersResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": I2C_COMMAND_NAMES[self.fields.header.code],
            "result": I2C_RESULT_NAMES[self.fields.header.result],
            "i2c_bus": I2cBus(self.fields.parameters.busId).name,
        }

# endregion

#================================================================================#
# region I2C SET PULL UP RESISTORS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class I2cSetPullUpResistorsRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
    			("pullUpValue", c_uint8)]

class I2cSetPullUpResistorsRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", I2cSetPullUpResistorsRequestParameters_t)]

I2cSetPullUpResistorsRequestArray_t = c_uint8 * sizeof(I2cSetPullUpResistorsRequestFields_t)

class I2cSetPullUpResistorsRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", I2cSetPullUpResistorsRequestArray_t ),
                ("fields", I2cSetPullUpResistorsRequestFields_t )]

# Response --------------------------------------------------------------------- #

class I2cSetPullUpResistorsResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8)]

class I2cSetPullUpResistorsResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", I2cSetPullUpResistorsResponseParameters_t)]

I2cSetPullUpResistorsResponseArray_t = c_uint8 * sizeof(I2cSetPullUpResistorsResponseFields_t)

class I2cSetPullUpResistorsResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", I2cSetPullUpResistorsResponseArray_t ),
                ("fields", I2cSetPullUpResistorsResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = I2cSetPullUpResistorsResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": I2C_COMMAND_NAMES[self.fields.header.code],
            "result": I2C_RESULT_NAMES[self.fields.header.result],
            "i2c_bus": I2cBus(self.fields.parameters.busId).name,
        }

# endregion

#================================================================================#
# region I2C CONTROLLER TRANSFER
#================================================================================#

MAX_I2C_TRANSFER_LENGTH = 1024

# Request ---------------------------------------------------------------------- #

class I2cControllerTransferRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
    			("transfLength", c_uint16),
                ("targetAddress", c_uint8),
                ("registerAddressLength", c_uint8),
                ("registerAddress", c_uint32),
                ("isNonStop", c_uint8)]

I2cControllerTransferRequestPayload_t = c_uint8 * MAX_I2C_TRANSFER_LENGTH

class I2cControllerTransferRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", I2cControllerTransferRequestParameters_t),
                ("payload", I2cControllerTransferRequestPayload_t)]

I2cControllerTransferRequestArray_t = c_uint8 * sizeof(I2cControllerTransferRequestFields_t)

class I2cControllerTransferRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", I2cControllerTransferRequestArray_t),
                ("fields", I2cControllerTransferRequestFields_t )]

    def toBytes(self) -> bytes:
        """
        Method to return an bytes object representing the command
        serialization.
        """
        length = sizeof(CommandRequestHeader_t) + sizeof(I2cControllerTransferRequestParameters_t)

        if self.fields.header.code == I2cCommandCodes.I2C_CONTROLLER_WRITE.value:
            length += self.fields.parameters.transfLength

        return bytes(self.data[:length])

# Response --------------------------------------------------------------------- #

class I2cControllerTransferResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
    			("payloadLength", c_uint16)]

I2cControllerTransferResponsePayload_t = c_uint8 * MAX_I2C_TRANSFER_LENGTH

class I2cControllerTransferResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", I2cControllerTransferResponseParameters_t),
                ("payload", I2cControllerTransferResponsePayload_t)]

I2cControllerTransferResponseArray_t = c_uint8 * sizeof(I2cControllerTransferResponseFields_t)

class I2cControllerTransferResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", I2cControllerTransferResponseArray_t ),
                ("fields", I2cControllerTransferResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        for i in range(len(data)):
            self.data[i] = data[i]

    def toDictionary(self) -> dict:
        response = {
            "id": self.fields.header.id,
            "command": I2C_COMMAND_NAMES[self.fields.header.code],
            "result": I2C_RESULT_NAMES[self.fields.header.result],
            "i2c_bus": I2cBus(self.fields.parameters.busId).name,
            "payload_length": self.fields.parameters.payloadLength,
        }

        if self.fields.header.code == I2cCommandCodes.I2C_CONTROLLER_READ.value:
            response["payload"] = self.fields.payload[:self.fields.parameters.payloadLength]
        
        return response

# endregion

#================================================================================#
# region I2C CONTROLLER SCAN BUS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class I2cControllerScanBusRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
                ("include10BitAddresses", c_uint8)]

class I2cControllerScanBusRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", I2cControllerScanBusRequestParameters_t)]
    
I2cControllerScanBusRequestArray_t = c_uint8 * sizeof(I2cControllerScanBusRequestFields_t)

class I2cControllerScanBusRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", I2cControllerScanBusRequestArray_t),
                ("fields", I2cControllerScanBusRequestFields_t)]
    
# Response --------------------------------------------------------------------- #

I2C_CONTROLLER_SCAN_BUS_PAYLOAD_LENGTH = 100

class I2cControllerScanBusResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("busId", c_uint8),
                ("count7BitAddresses", c_uint8),
                ("count10BitAddresses", c_uint8)]

I2cControllerScanBusResponsePayload_t = c_uint8 * I2C_CONTROLLER_SCAN_BUS_PAYLOAD_LENGTH

class I2cControllerScanBusResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", I2cControllerScanBusResponseParameters_t),
                ("payload", I2cControllerScanBusResponsePayload_t)]

I2cControllerScanBusResponseArray_t = c_uint8 * sizeof(I2cControllerScanBusResponseFields_t)

class I2cControllerScanBusResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", I2cControllerScanBusResponseArray_t),
                ("fields", I2cControllerScanBusResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        for i in range(len(data)):
            self.data[i] = data[i]

    def __formatDetected10BitAddresses(self, rawData, count):
        """
        This function formats the byte list of detected 10-bit addresses into a list of uint16_t.
        """
        addresses = []
        for i in range(count):
            address = (rawData[2*i + 1] << 8) | rawData[2*i]
            addresses.append(address)
        return addresses

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": I2C_COMMAND_NAMES[self.fields.header.code],
            "result": I2C_RESULT_NAMES[self.fields.header.result],
            "i2c_bus": I2cBus(self.fields.parameters.busId).name,
            "detected_7_bit_addresses": self.fields.payload[:self.fields.parameters.count7BitAddresses],
            "detected_10_bit_addresses": self.__formatDetected10BitAddresses(self.fields.payload[self.fields.parameters.count7BitAddresses:], self.fields.parameters.count10BitAddresses)
        }

# endregion


from ..common_definitions import *

#================================================================================#
# region SPI COMMAND DEFINITIONS
#================================================================================#

class SpiCommandCodes(Enum):
    """
    Enumeration of the SPI command codes.
    """
    SPI_DEINIT                      = makeCommandCode(Group.SPI.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 1)
    SPI_CONTROLLER_INIT             = makeCommandCode(Group.SPI.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 1)
    SPI_CONTROLLER_SET_PARAMETERS   = makeCommandCode(Group.SPI.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 2)
    SPI_CONTROLLER_GET_PARAMETERS   = makeCommandCode(Group.SPI.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 3)
    SPI_CONTROLLER_TRANSFER         = makeCommandCode(Group.SPI.value, CommandRole.CONTROLLER.value, CommandType.REQUEST_RESPONSE.value, 4)

SPI_COMMAND_NAMES = {
    result.value: result.name.replace("_", " ") for result in SpiCommandCodes
}

# endregion

#================================================================================#
# region SPI COMMON DEFINITIONS
#================================================================================#

# SPI Bus Configuration
SPI_CONTROLLER_MIN_FREQUENCY = 10000       # In Hz -> 10 kHz
SPI_CONTROLLER_MAX_FREQUENCY = 50000000    # In Hz -> 50 MHz

class SpiControllerBitOrder(Enum):
    MSB    = 0x00
    LSB    = 0x01

class SpiControllerMode(Enum):
    MODE_0    = 0x00
    MODE_1    = 0x01
    MODE_2    = 0x02
    MODE_3    = 0x03

class SpiControllerDataWidth(Enum):
    _8_BITS_DATA    = 0x00
    _16_BITS_DATA   = 0x01

class SpiControllerChipSelect(Enum):
    CHIP_SELECT_0   = 0x00
    CHIP_SELECT_1   = 0x01
    CHIP_SELECT_2   = 0x02
    CHIP_SELECT_3   = 0x03

class SpiControllerChipSelectPolarity(Enum):
    ACTIVE_LOW  = 0x00
    ACTIVE_HIGH = 0x01

class SpiControllerConfigurationParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("bitOrder", c_uint8),
                ("mode", c_uint8),
                ("dataWidth", c_uint8),
                ("chipSelect", c_uint8),
                ("chipSelectPol", c_uint8),
                ("frequency", c_uint32)]

# endregion

#================================================================================#
# region SPI CONTROLLER INIT
#================================================================================#

# Request ---------------------------------------------------------------------- #

class SpiControllerInitRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", SpiControllerConfigurationParameters_t)]

SpiControllerInitRequestArray_t = c_uint8 * sizeof(SpiControllerInitRequestFields_t)

class SpiControllerInitRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", SpiControllerInitRequestArray_t ),
                ("fields", SpiControllerInitRequestFields_t )]

# Response --------------------------------------------------------------------- #

class SpiControllerInitResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

SpiControllerInitResponseArray_t = c_uint8 * sizeof(SpiControllerInitResponseFields_t)

class SpiControllerInitResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", SpiControllerInitResponseArray_t ),
                ("fields", SpiControllerInitResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = SpiControllerInitResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": SPI_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region SPI CONTROLLER SET PARAMETERS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class SpiControllerSetParameterRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", SpiControllerConfigurationParameters_t)]

SpiControllerSetParameterRequestArray_t = c_uint8 * sizeof(SpiControllerSetParameterRequestFields_t)

class SpiControllerSetParameterRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", SpiControllerSetParameterRequestArray_t ),
                ("fields", SpiControllerSetParameterRequestFields_t )]

# Response --------------------------------------------------------------------- #

class SpiControllerSetParameterResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

# Union array
SpiControllerSetParameterResponseArray_t = c_uint8 * sizeof(SpiControllerSetParameterResponseFields_t)

class SpiControllerSetParameterResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", SpiControllerSetParameterResponseArray_t ),
                ("fields", SpiControllerSetParameterResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = SpiControllerSetParameterResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": SPI_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region SPI CONTROLLER TRANSFER
#================================================================================#

MAX_SPI_TRANSFER_LENGTH = 1024

# Request ---------------------------------------------------------------------- #

class SpiControllerTransferRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("payloadLength", c_uint16),
                ("transferLength", c_uint16)]

SpiControllerTransferRequestPayload_t = c_uint8 * MAX_SPI_TRANSFER_LENGTH

class SpiControllerTransferRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", SpiControllerTransferRequestParameters_t),
                ("payload", SpiControllerTransferRequestPayload_t)]

SpiControllerTransferRequestArray_t = c_uint8 * sizeof(SpiControllerTransferRequestFields_t)

class SpiControllerTransferRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", SpiControllerTransferRequestArray_t ),
                ("fields", SpiControllerTransferRequestFields_t )]

    def toBytes(self) -> bytes:
        """
        Method to return an bytes object representing the command
        serialization.
        """
        length = sizeof(CommandRequestHeader_t) + sizeof(SpiControllerTransferRequestParameters_t) + self.fields.parameters.payloadLength
        return bytes(self.data[:length])

# Response --------------------------------------------------------------------- #

class SpiControllerTransferResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("payloadLength", c_uint16)]

SpiControllerTransferResponsePayload_t = c_uint8 * MAX_SPI_TRANSFER_LENGTH

class SpiControllerTransferResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", SpiControllerTransferResponseParameters_t),
                ("payload", SpiControllerTransferResponsePayload_t)]

SpiControllerTransferResponseArray_t = c_uint8 * sizeof(SpiControllerTransferResponseFields_t)

class SpiControllerTransferResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", SpiControllerTransferResponseArray_t ),
                ("fields", SpiControllerTransferResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        for i in range(len(data)):
            self.data[i] = data[i]

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": SPI_COMMAND_NAMES[self.fields.header.code],
            "result": COMMON_RESULT_NAMES[self.fields.header.result],
            "payload_length": self.fields.parameters.payloadLength,
            "payload": self.fields.payload[:self.fields.parameters.payloadLength]
        }

# endregion


from setuptools import setup, find_packages

setup(
    name='binhopulsar',
    version='1.1.0',
    packages=find_packages(),
    description='Basic SDK for Binho Pulsar',
    long_description=open('README.md', encoding="utf8").read(),
    long_description_content_type='text/markdown',
    author='Binho LLC',
    author_email='support@binho.io',
    url='https://github.com/binhollc/PulsarSDK',
    license='Private',
    install_requires=[
        "pyserial",
        "hidapi",
        "psutil",
    ],
    package_data= {
        'binhopulsar': ['supernovasdk/binhosupernova/*',
                        'supernovasdk/binhosupernova/**/*'],
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)


import threading
import inspect
import hid

from ..utils.system_message import SystemMessage, SystemOpcode, SystemModules
from .usb_transfer_protocol import USB_FS_INTERRUPT_ENDPOINT_SIZE, USB_HS_INTERRUPT_ENDPOINT_SIZE, UsbTransferProtocol

#================================================================================#
# region Constants definitions
#================================================================================#

ENDPOINT_ID = 0x00

# VID and PID constants definition.
BINHO_VID       = 0x1FC9
SUPERNOVA_PID   = 0x82FC
PULSAR_PID      = 0x82FD

# USB transfer error opcode returned by hidapi write method.
USB_TRANSFER_ERROR_OPCODE = -1

#endregion


#================================================================================#
# region USB HID MANAGER class
#================================================================================#

class UsbHidManager:
    """
    This class is responsible for the management of the USB HID communication. It exposes
    the method to open and close the connections, as well as the method to send data over USB
    to the host adapter. Additionally, it constantly listens to the reception of data and invokes
    the onReceiveCallback every time data is received.

    Attributes
    ----------
    binhoDeviceList: dict
        Class attribute to hold the different Binho devices characteristics.
    thread: threading.Thread
        Thread that is in charge of the data reception.
    run: bool
        Boolean flag used as the variable of control of the thread.
    vendorId: int
        USB Vendor ID
    productId: int
        USB Product ID
    usbHidDevice: hid.device
        Instance of the hid device class which connects withe the low-level HID driver.
    endpointSize: int
        USB Interrupt transfers endpoint size.
    onReceiveCallback:
        Callback function invoked when a new response or notification is received to pass
        the data to the host application.

    """

    # Class attribute.
    binhoDeviceList = {
        SUPERNOVA_PID: USB_FS_INTERRUPT_ENDPOINT_SIZE,
        PULSAR_PID: USB_HS_INTERRUPT_ENDPOINT_SIZE
    }

    @staticmethod
    def enumerate(productId):
        # Get a list of dictionaries of devices info.
        devices = hid.enumerate(BINHO_VID, productId)

        # For each device, print de VID and PID in hex, and convert the path
        # from bytes to String
        for device in devices:
            device['path'] = device['path'].decode(encoding='utf-8')
            device['vendor_id'] = hex(device['vendor_id'])
            device['product_id'] = hex(device['product_id'])

        # Return list of devices.
        return devices

    def __init__(self, productId):

        # Threading
        self.thread = None                                              # Reception thread.
        self.run = False                                                # Variable of thread control.

        # USB specs
        self.vendorId = BINHO_VID                                       # USB Vendor ID.
        self.productId = productId                                      # USB Product ID.
        self.endpointSize = UsbHidManager.binhoDeviceList[productId]    # USB Interrupt transfer endpoint size.
        self.usbHidDevice = hid.device()                                # Instance of hid.device()

        self.transferProtocol = UsbTransferProtocol(self.endpointSize)

        # Responses handling
        self.onReceiveCallback = None                                   # Callback to pass the response/notification received.

    # Private methods --------------------------------------------------------------------------------- #

    def __isRunning(self):
        """
        This method returns True if the reception thread is alive and running.
        """
        return self.thread and self.thread.is_alive() and self.run

    def __stop(self):
        """
        This method ends the USB receiver process and deletes the instance of thread
        so that is possible to create a new instance to reconnect the communication since
        it is not possible to invoke threading.Thread.start() method more than once.
        """
        self.run = False
        if self.thread:
            self.thread.join()
            self.thread = None

    def __start(self):
        """
        This method starts the Thread target function. If a thread is already
        created and running, first ends the current process and starts a new instance since
        it is not possible to invoke threading.Thread.start() method more than once.
        """
        if self.__isRunning():
            self.__stop()

        self.run = True
        self.thread = threading.Thread(target = self.__main, name='USB Receiver', daemon=True)
        self.thread.start()

    def __main(self):
        """
        This method is the Thread target function.
        """
        while self.run == True:
            try:
                # Block for 100 ms to wait for a new message receive in the USB port from the Supernova.
                usbHostAdapterMessage = bytes(self.usbHidDevice.read(self.endpointSize, 100))

                # If new data was received, pass the packet to the USB Transfer Protocol and check if the transfer is complete.
                if usbHostAdapterMessage:
                    transfer = self.transferProtocol.receiveUsbInputTransferPackets(usbHostAdapterMessage)
                    if transfer and self.onReceiveCallback:
                        self.onReceiveCallback(transfer, None)

            except OSError as e:     # This exception is raised from self.usbHostAdapter.read when the Supernova is removed.
                # Create a custom error.
                error = SystemMessage(SystemModules.SYSTEM, SystemOpcode.UNEXPECTED_DISCONNECTION, f"Error {SystemOpcode.UNEXPECTED_DISCONNECTION.name}: Unexpected Supernova disconnection.")
                # Notify to the client.
                self.onReceiveCallback(None, error)
                # Kill process
                self.run = False

    # Public methods --------------------------------------------------------------------------------- #

    def open(self, serial = None, path:str = None):
        """
        This method is in charge of opening the connection with the USB Host Adapter through the HDI API library.

        Parameters
        ----------
        serial: int
            String representing the serial number of the USB Host Adapter.
        path: str
            String representing the OS path of the USB Host Adapter.

        Returns
        -------
        dict:
            A python dictionary representation of a SystemMessage instance. The opcode of the message is:
            - SystemOpcode.OK if the connection with the HID device is completed successfully.
            - SystemOpcode.OPEN_CONNECTION_FAIL if the connection of the HID instance fails.
        """
        # Create response instance.
        sys_response = SystemMessage(SystemModules.SYSTEM, SystemOpcode.OK, "Connection with Supernova device opened successfully.")

        try:
            # Close the connection before reopening it.
            # Stop and kill thread.
            if self.__isRunning():
                self.__stop()

            # Close HID device.
            self.usbHidDevice.close()

            # Open HID connection.
            if path:
                path_bytes = path.encode(encoding='utf-8')
                self.usbHidDevice.open_path(path_bytes)
            else:
                self.usbHidDevice.open(self.vendorId, self.productId, serial)

            # Start thread.
            self.__start()

        except Exception as exc:
            sys_response.opcode = SystemOpcode.OPEN_CONNECTION_FAIL
            sys_response.message = f"Open connection failed. Exception type: {type(exc)}. Exception message: {exc}."

        return sys_response.toDictionary()

    def isRunning(self):
        """
        This method returns True if the USB connection is open and working.
        """
        return self.__isRunning()

    def close(self):
        """
        This method is in charge of closing the connection with the USB Host Adapter through the HDI API library.

        Returns
        -------
        dict:
            A python dictionary representation of a SystemMessage instance. The opcode of the message is:
            - SystemOpcode.OK if the connection with the HID device is closed successfully.
            - SystemOpcode.OPEN_CONNECTION_REQUIRED if the was not previously opened.

        """

        sys_response = SystemMessage(SystemModules.SYSTEM, SystemOpcode.OK, "Communication closed successfully.")

        if self.__isRunning():
            self.__stop()
            # Close HID device.
            self.usbHidDevice.close()
        else:
            sys_response.opcode = SystemOpcode.OPEN_CONNECTION_REQUIRED
            sys_response.message = "It is required to open connection with a Supernova first. Invoke open() method."

        return sys_response.toDictionary()

    def join(self):
        """
        This method waits until the reception thread ends.
        """
        self.thread.join()

    def setOnReceiveCallback(self, callback):
        """
        This function sets the USB receiver callback function.

        Parameters
        ----------
        callback:
            2 parameters signature callback function.

        Returns
        -------
        dict:
            A python dictionary representation of a SystemMessage instance. The opcode of the message is:
            - SystemOpcode.OK if the callback definition is correct and it is assigned to the onReceiveCallback attribute.
            - SystemOpcode.INVALID_CALLBACK_SIGNATURE if the signature of the function is not correct.

        """

        sys_response = SystemMessage(SystemModules.SYSTEM, SystemOpcode.OK, "On event callback function registered successfully.")

        # Get the function signature
        func_signature = inspect.signature(callback)

        # Get the parameters of the function
        func_parameters = func_signature.parameters

        # Check if the function has exactly 2 parameters
        if len(func_parameters) != 2:
            sys_response.opcode = SystemOpcode.INVALID_CALLBACK_SIGNATURE
            sys_response.message = "The function must accept 2 arguments: callback(supernova_response, system_message)."
        else:
            # Save the callback function
            self.onReceiveCallback = callback

        return sys_response.toDictionary()

    def send(self, transfer):
        """
        This method is in charge of sending data to the USB Host Adapter through the HDI API library.

        Parameters
        ----------
        transfers: list
            List of chunks of data to be sent to the USB Host Adapter over USB.

        Returns
        -------
        int:
            An integer representing the result of the USB transfer. The value USB_TRANSFER_ERROR_OPCODE
            represents the transfer failure.

        """

        packets = self.transferProtocol.createUsbOutputTransferPackets(transfer)
        result = 0

        for packet in packets:
            # Prepend ENDPOINT_ID to the transfer
            transfer_with_endpoint = bytes([ENDPOINT_ID]) + packet
            # Send request to USB host adapter
            result = self.usbHidDevice.write(transfer_with_endpoint)

            if result == USB_TRANSFER_ERROR_OPCODE:
                break
        
        return result

#endregion